from __future__ import annotations

import csv
import datetime
from typing import TYPE_CHECKING, Any, Iterable

from flask import Blueprint

import ckan.authz as authz
import ckan.plugins.toolkit as tk
from ckan import model
from ckan.common import streaming_response
from ckan.lib.helpers import Page

if TYPE_CHECKING:
    from ckan.types import Context

CONFIG_BASE_TEMPLATE = "ckanext.check_link.report.base_template"
DEFAULT_BASE_TEMPLATE = "check_link/base_admin.html"

CONFIG_REPORT_URL = "ckanext.check_link.report.url"
DEFAULT_REPORT_URL = "/check-link/report/global"

CSV_COLUMNS = [
    "Data Record title",
    "Data Resource title",
    "Organisation",
    "State",
    "Error type",
    "Link to Data resource",
    "Date and time checked",
]

bp = Blueprint("check_link", __name__)


def get_blueprints():
    report_url = tk.config.get(CONFIG_REPORT_URL, DEFAULT_REPORT_URL)
    if report_url:
        bp.add_url_rule(report_url, view_func=report)

    return [bp]


def report():
    if not authz.is_authorized_boolean(
        "check_link_view_report_page", {"user": tk.g.user}, {}
    ):
        return tk.abort(403)

    params = {
        "attached_only": True,
        "exclude_state": ["available"],
    }

    fmt = tk.request.args.get("format")
    if fmt == "csv":
        resp = streaming_response(
            _stream_csv(
                _iterate_resuts("check_link_report_search", params, {"user": tk.g.user})
            ),
            mimetype="text/csv",
            with_context=True,
        )
        today = datetime.date.today()
        resp.headers[
            "content-disposition"
        ] = f'attachment; filename="VPSDDLinkReport-{today:%d%m%Y}.csv"'
        return resp

    try:
        page = max(1, tk.asint(tk.request.args.get("page", 1)))
    except ValueError:
        page = 1

    per_page = 10
    reports = tk.get_action("check_link_report_search")(
        {},
        dict(params, limit=per_page, offset=per_page * page - per_page),
    )

    def pager_url(*args: Any, **kwargs: Any):
        return tk.url_for("check_link.report", **kwargs)

    base_template = tk.config.get(CONFIG_BASE_TEMPLATE, DEFAULT_BASE_TEMPLATE)
    return tk.render(
        "check_link/report.html",
        {
            "base_template": base_template,
            "page": Page(
                reports["results"],
                url=pager_url,
                page=page,
                item_count=reports["count"],
                items_per_page=per_page,
                presliced_list=True,
            ),
        },
    )


class _FakeBuffer:
    def write(self, value):
        return value


def _stream_csv(reports):
    writer = csv.writer(_FakeBuffer())

    yield writer.writerow(CSV_COLUMNS)
    _org_cache = {}

    for report in reports:
        owner_org = report["details"]["package"]["owner_org"]
        if owner_org not in _org_cache:
            _org_cache[owner_org] = model.Group.get(owner_org)

        yield writer.writerow(
            [
                report["details"]["package"]["title"],
                report["details"]["resource"]["name"] or "Unknown",
                _org_cache[owner_org] and _org_cache[owner_org].title,
                report["state"],
                report["details"]["explanation"],
                tk.url_for(
                    "resource.read",
                    id=report["package_id"],
                    resource_id=report["resource_id"],
                    _external=True,
                ),
                tk.h.render_datetime(report["created_at"], None, True),
            ]
        )


def _iterate_resuts(
    action: str,
    params: dict[str, Any],
    context: Context | None = None,
    offset: int = 0,
    chunk_size: int = 10,
) -> Iterable[dict[str, Any]]:
    while True:
        result = tk.get_action(action)(
            context or {},
            dict(params, limit=chunk_size, offset=offset),
        )
        yield from result["results"]
        offset += chunk_size
        if offset >= result["count"]:
            break



from __future__ import annotations

import logging
from collections import Counter
from itertools import islice
from typing import Iterable, TypeVar

import click

import ckan.model as model
import ckan.plugins.toolkit as tk

from .model import Report

T = TypeVar("T")
log = logging.getLogger(__name__)


def get_commands():
    return [check_link]


@click.group(short_help="Check link availability")
def check_link():
    pass


@check_link.command()
@click.option(
    "-d", "--include-draft", is_flag=True, help="Check draft packages as well"
)
@click.option(
    "-p", "--include-private", is_flag=True, help="Check private packages as well"
)
@click.option(
    "-c",
    "--chunk",
    help="Number of packages that processed simultaneously",
    default=1,
    type=click.IntRange(
        1,
    ),
)
@click.option(
    "-d", "--delay", default=0, help="Delay between requests", type=click.FloatRange(0)
)
@click.option(
    "-t", "--timeout", default=10, help="Request timeout", type=click.FloatRange(0)
)
@click.argument("ids", nargs=-1)
def check_packages(
    include_draft: bool,
    include_private: bool,
    ids: tuple[str, ...],
    chunk: int,
    delay: float,
    timeout: float,
):
    """Check every resource inside each package.

    Scope can be narrowed via arbitary number of arguments, specifying
    package's ID or name.

    """
    user = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    context = {"user": user["name"]}

    check = tk.get_action("check_link_search_check")
    states = ["active"]

    if include_draft:
        states.append("draft")

    q = model.Session.query(model.Package.id).filter(
        model.Package.state.in_(states),
    )

    if not include_private:
        q = q.filter(model.Package.private == False)

    if ids:
        q = q.filter(model.Package.id.in_(ids) | model.Package.name.in_(ids))

    stats = Counter()
    with click.progressbar(q, length=q.count()) as bar:
        while True:
            buff = _take(bar, chunk)
            if not buff:
                break

            result = check(
                context.copy(),
                {
                    "fq": "id:({})".format(" OR ".join(p.id for p in buff)),
                    "save": True,
                    "clear_available": True,
                    "include_drafts": include_draft,
                    "include_private": include_private,
                    "skip_invalid": True,
                    "rows": chunk,
                    "link_patch": {"delay": delay, "timeout": timeout},
                },
            )
            stats.update(r["state"] for r in result)
            overview = (
                ", ".join(
                    f"{click.style(k,  underline=True)}:"
                    f" {click.style(str(v),bold=True)}"
                    for k, v in stats.items()
                )
                or "not available"
            )
            bar.label = f"Overview: {overview}"
            bar.update(chunk)

    click.secho("Done", fg="green")


def _take(seq: Iterable[T], size: int) -> list[T]:
    return list(islice(seq, size))


@check_link.command()
@click.option(
    "-d", "--delay", default=0, help="Delay between requests", type=click.FloatRange(0)
)
@click.option(
    "-t", "--timeout", default=10, help="Request timeout", type=click.FloatRange(0)
)
@click.argument("ids", nargs=-1)
def check_resources(ids: tuple[str, ...], delay: float, timeout: float):
    """Check every resource on the portal.

    Scope can be narrowed via arbitary number of arguments, specifying
    resource's ID or name.
    """
    user = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    context = {"user": user["name"]}

    check = tk.get_action("check_link_resource_check")
    q = model.Session.query(model.Resource.id).filter_by(state="active")
    if ids:
        q = q.filter(model.Resource.id.in_(ids))

    stats = Counter()
    total = q.count()
    overview = "Not ready yet"
    with click.progressbar(q, length=total) as bar:
        for res in bar:
            bar.label = f"Current: {res.id}. Overview({total} total): {overview}"
            try:
                result = check(
                    context.copy(),
                    {
                        "save": True,
                        "clear_available": True,
                        "id": res.id,
                        "link_patch": {"delay": delay, "timeout": timeout},
                    },
                )
            except tk.ValidationError as e:
                log.error("Cannot check %s: %s", res.id, e)
                result = {"state": "exception"}

            stats[result["state"]] += 1
            overview = (
                ", ".join(
                    f"{click.style(k,  underline=True)}:"
                    f" {click.style(str(v),bold=True)}"
                    for k, v in stats.items()
                )
                or "not available"
            )
            bar.label = f"Current: {res.id}. Overview({total} total): {overview}"

    click.secho("Done", fg="green")


@check_link.command()
@click.option(
    "-o",
    "--orphans-only",
    is_flag=True,
    help="Only drop reports that point to an unexisting resource",
)
def delete_reports(orphans_only: bool):
    """Delete check-link reports."""
    q = model.Session.query(Report)
    if orphans_only:
        q = q.outerjoin(model.Resource, Report.resource_id == model.Resource.id).filter(
            Report.resource_id.isnot(None),
            model.Resource.id.is_(None) | (model.Resource.state != "active"),
        )

    user = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    context = {"user": user["name"]}

    action = tk.get_action("check_link_report_delete")
    with click.progressbar(q, length=q.count()) as bar:
        for report in bar:
            action(context.copy(), {"id": report.id})



from unittest.mock import ANY

import pytest

import ckan.plugins.toolkit as tk

# from aioresponses import aioresponses
from ckan.tests.helpers import call_action


@pytest.fixture
def rmock(httpx_mock):
    return httpx_mock


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestUrl:
    def test_allow_multi_url_validatoin(self, faker, rmock):
        url1 = faker.url()
        url2 = faker.url()

        # rmock.head(url1, status=200)
        # rmock.head(url2, status=404)
        rmock.add_response(url=url1, status_code=200, method="HEAD")
        rmock.add_response(url=url2, status_code=404, method="HEAD")

        result = call_action("check_link_url_check", url=[url1, url2])

        assert result == [
            {
                "code": 200,
                "explanation": ANY,
                "reason": ANY,
                "state": "available",
                "url": url1,
            },
            {
                "code": 404,
                "explanation": ANY,
                "reason": ANY,
                "state": "missing",
                "url": url2,
            },
        ]

    def test_not_saved_by_defaut(self, faker, rmock):
        url = faker.url()

        # rmock.head(url, status=200)
        rmock.add_response(url=url, status_code=200, method="HEAD")

        call_action("check_link_url_check", url=url)

        with pytest.raises(tk.ObjectNotFound):
            call_action("check_link_report_show", url=url)

    def test_can_be_saved(self, faker, rmock):
        url = faker.url()

        # rmock.head(url, status=200)
        rmock.add_response(url=url, status_code=200, method="HEAD")

        call_action("check_link_url_check", url=url, save=True)
        report = call_action("check_link_report_show", url=url)
        assert report == {
            "created_at": ANY,
            "details": {
                "code": 200,
                "explanation": "Link is available",
                "reason": "OK",
            },
            "id": ANY,
            "resource_id": None,
            "package_id": None,
            "state": "available",
            "url": url,
        }


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestResource:
    def test_not_saved_by_defaut(self, resource, rmock):
        # rmock.head(resource["url"], status=200)
        rmock.add_response(url=resource["url"], status_code=200, method="HEAD")
        call_action("check_link_resource_check", id=resource["id"])

        with pytest.raises(tk.ObjectNotFound):
            call_action("check_link_report_show", resource_id=resource["id"])

    def test_can_be_saved(self, resource, rmock):
        # rmock.head(resource["url"], status=200)
        rmock.add_response(url=resource["url"], status_code=200, method="HEAD")
        call_action("check_link_resource_check", id=resource["id"], save=True)
        report = call_action("check_link_report_show", resource_id=resource["id"])
        assert report["resource_id"] == resource["id"]
        assert report["package_id"] == resource["package_id"]


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestPackage:
    def test_basic(self, resource_factory, rmock, package):
        resource = resource_factory(package_id=package["id"])
        # rmock.head(resource["url"], status=200)
        rmock.add_response(url=resource["url"], status_code=200, method="HEAD")
        result = call_action("check_link_package_check", id=package["id"])
        assert len(result) == 1
        assert result[0]["code"] == 200

    def test_empty(self, package):
        result = call_action("check_link_package_check", id=package["id"])
        assert result == []



from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    String,
    UnicodeText,
    UniqueConstraint,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import backref, relationship
from typing_extensions import Self

import ckan.model as model
from ckan.lib.dictization import table_dictize
from ckan.lib.dictization.model_dictize import resource_dictize
from ckan.model.types import make_uuid

from .base import Base


class Report(Base):
    __tablename__ = "check_link_report"

    id = Column(UnicodeText, primary_key=True, default=make_uuid)
    url = Column(UnicodeText, nullable=False)
    state = Column(String(20), nullable=False)

    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)

    resource_id = Column(
        UnicodeText, ForeignKey(model.Resource.id), nullable=True, unique=True
    )

    details = Column(JSONB, nullable=False, default=dict)

    package_id = association_proxy("resource", "package_id")
    package = association_proxy("resource", "package")

    resource = relationship(
        model.Resource,
        backref=backref(
            "check_link_report", cascade="all, delete-orphan", uselist=False
        ),
    )

    UniqueConstraint(url, resource_id)

    def touch(self):
        self.created_at = datetime.utcnow()

    def dictize(self, context: dict[str, Any]) -> dict[str, Any]:
        result = table_dictize(self, context, package_id=self.package_id)

        if context.get("include_resource") and self.resource_id:
            result["details"]["resource"] = resource_dictize(self.resource, context)

        if context.get("include_package") and self.package_id:
            result["details"]["package"] = resource_dictize(self.package, context)

        return result

    @classmethod
    def by_resource_id(cls, id_: str) -> Optional[Self]:
        if not id_:
            return

        return model.Session.query(cls).filter(cls.resource_id == id_).one_or_none()

    @classmethod
    def by_url(cls, url: str) -> Optional[Self]:
        return (
            model.Session.query(cls)
            .filter(cls.resource_id.is_(None), cls.url == url)
            .one_or_none()
        )



from __future__ import annotations

import ckan.plugins.toolkit as tk

from ckanext.toolbelt.decorators import Collector

CONFIG_HEADER_LINK = "ckanext.check_link.show_header_link"
DEFAULT_HEADER_LINK = False

helper, get_helpers = Collector("check_link").split()


@helper
def show_header_link() -> bool:
    return tk.asbool(tk.config.get(CONFIG_HEADER_LINK, DEFAULT_HEADER_LINK))



import pytest

from ckan.plugins import plugin_loaded


@pytest.mark.ckan_config("ckan.plugins", "check_link")
@pytest.mark.usefixtures("with_plugins")
def test_plugin():
    assert plugin_loaded("check_link")
